import urllib.parse


class QueryBuilder:

    def __init__(self):
        self._query = None
        self._entity = None
        self._action = None
        self._fetchxml = None
        self._selects = None
        self._filters = []
        self._expand = None
        self._order = []
        self._top = None
        self._skip = None

    def entity(self, entity: str):
        if not isinstance(entity, (str)):
            raise Exception('Must pass a string to entity()')
        self._entity = entity
        return self

    def action(self, action: str):
        if not isinstance(action, (str)):
            raise Exception('Must pass a string to action()')
        self._action = action
        return self

    def select(self, l: list):
        if not isinstance(l, (list)):
            raise Exception('Must pass a list to select()')
        self._selects = l
        return self

    def filter(self, field=None, comp=None, value=None, type=None, logop='and'):
        if field is None or comp is None or value is None or type is None:
            raise Exception('Must set field, comp, value and type')

        if logop not in ['and', 'or', 'and not']:
            raise Exception('Logical operator must be one of: and/or/and-not')

        if comp not in ['eq', 'ne', 'gt', 'ge', 'lt', 'le', 'startswith', 'in', 'contains', 'between', 'notin']:
            raise Exception('Invalid comparison type')

        self._filters.append({
            'by': field,
            'comp': comp,
            'value': value,
            'type': type,
            'logop': logop
        })

        return self

    def expand(self, l: list):
        if not isinstance(l, (list)):
            raise Exception('Must pass a list to expand()')
        self._expand = l
        return self

    def order(self, by: list, mode: str):
        if not isinstance(by, (list)):
            raise Exception('Must pass a list to order() as by')

        if mode not in ['asc', 'desc']:
            raise Exception('Mode must be one of asc or desc')

        self._order.append({
            'by': by,
            'mode': mode
        })

        return self

    def limit(self, skip: int, top: int):
        if not isinstance(skip, (int)):
            raise Exception('Must pass an int to limit() as skip')
        if not isinstance(top, (int)):
            raise Exception('Must pass an int to limit() as top')
        self._top = top
        if skip > 0:
            self._skip = skip
        return self

    def fetchxml(self, xml: str):
        if self._action is not None or self._action is not None or len(self._filters) > 0:
            raise Exception('FetchXML can not be used in conjunction with other query methods')

        self._fetchxml = xml
        return self

    def _buildentity(self):
        if self._entity is None:
            raise Exception('Entity must be set')

        self._query = self._entity

    def _buildaction(self):
        self._query += '/Microsoft.Dynamics.CRM.' + self._action;

    def _buildfetchxml(self):
        self._query += 'fetchXml='+urllib.parse.quote_plus(self._fetchxml)

    def _buildselects(self):
        self._query += '$select='
        self._query += ','.join(self._selects)

    def _buildfilters(self):
        if self._query[-1:] != "?":
            self._query += '&'

        self._query += "$filter="
        _first = True
        ftext = ""
        for filter in self._filters:
            if _first is not True:
                ftext += " "+filter['logop']+" "
            _first = False

            if filter['comp'] == 'startswith':
                ftext += "startswith("+filter['by']+", '"+urllib.parse.quote_plus(str(filter['value']))+"')"
            elif filter['comp'] == 'between':
                ftext += "Microsoft.Dynamics.CRM.Between(PropertyName='"+filter['by']+"', PropertyValues=['"+urllib.parse.quote_plus(str(filter['value'][0]))+"','"+urllib.parse.quote_plus(str(filter['value'][1]))+"'])"
            elif filter['comp'] == 'in':
                ftext += "Microsoft.Dynamics.CRM.In(PropertyName='"+filter['by']+"',PropertyValues=['"+"','".join(filter['value'])+"'])"
            elif filter['comp'] == 'notin':
                ftext += "Microsoft.Dynamics.CRM.NotIn(PropertyName='"+filter['by']+"',PropertyValues=['"+"','".join(filter['value'])+"'])"
            elif filter['comp'] == 'contains':
                ftext += "contains("+filter['by']+", '"+urllib.parse.quote_plus(filter['value'])+"')"
            else:
                ftext += filter['by'] + " " + filter['comp'] + " "
                if filter['type'] == 'str':
                    ftext += "'"+urllib.parse.quote_plus(filter['value'])+"'"
                elif filter['type'] == 'bool':
                    if filter['value'] is True:
                        ftext += "true"
                    else:
                        ftext += "false"
                else:
                    ftext += filter['value']

            self._query += urllib.parse.quote_plus(ftext)

    def _buildexpand(self):
        if self._query[-1:] != "?":
            self._query += '&'

        self._query += 'expand='
        self._query += ','.join(self._expand)

    def _buildorder(self):
        if self._query[-1:] != "?":
            self._query += '&'

        self._query += "$orderby="
        first = True
        for o in self._order:
            if first is False:
                self._query += ","
            self._query += ",".join(o['by'])+" "+o['mode']
            first = False

    def _buildlimit(self):
        if self._query[-1:] != "?":
            self._query += '&'

        self._query += "$top="+str(self._top)

        if self._skip is not None:
            self._query += "&$skip="+str(self._skip)

    def buildquery(self):
        self._buildentity()
        if self._action is not None:
            self._buildaction()
        self._query += '?'

        if self._fetchxml is not None:
            self._buildfetchxml()
            return self._query

        if self._selects is not None:
            self._buildselects()

        if len(self._filters) > 0:
            self._buildfilters()

        if self._expand is not None:
            self._buildexpand()

        if len(self._order) > 0:
            self._buildorder()

        if self._top is not None:
            self._buildlimit()

        return self._query



import requests
import pydynamics


class Client:

    def __init__(self, token, endpoint):
        self._token = str(token)
        self._endpoint = endpoint
        self._headers = {'Accept': 'application/json', 'Content-type': 'application/json', 'OData-Version': '4.0',
                         'OData-MaxVersion': '4.0', 'prefer': 'odata.include-annotations=*',
                         'Authorization': 'Bearer '+self._token}

    def select(self, query):
        if not isinstance(query, pydynamics.querybuilder.QueryBuilder):
            raise Exception('Query must be of type QueryBuilder')

        resp = requests.get(self._endpoint + query.buildquery(), headers=self._headers)
        results = resp.json()['value']

        nextlink = None
        if '@odata.nextLink' in resp.json():
            nextlink = resp.json()['@odata.nextLink']

        while nextlink is not None:
            resp = requests.get(nextlink, headers=self._headers)
            results = results + resp.json()['value']
            if '@odata.nextLink' in resp.json():
                nextlink = resp.json()['@odata.nextLink']
            else:
                nextlink = None


        return results


import requests
from lxml import etree
from uuid import uuid4
import datetime


def get(url, username, password):
    try:
        crmres = requests.get(url + "XrmServices/2011/Organization.svc?wsdl=wsdl0")
    except requests.exceptions.RequestException as e:
        raise Exception(e)

    try:
        adfs = etree.fromstring(crmres.text)
    except ImportError:
        raise Exception('Failed to import XML response')

    ns = {'ms-xrm': 'http://schemas.microsoft.com/xrm/2011/Contracts/Services'}
    ident = adfs.findall('.//ms-xrm:Identifier', ns)
    if len(ident) == 0:
        raise Exception('Could not find ADFS URL in XML response')

    adfsurl = ident[0].text.replace('http:', 'https:')
    urnaddress = url + "XRMServices/2011/Organization.svc"
    usernamemixed = adfsurl + "/13/usernamemixed"

    xml = "<s:Envelope xmlns:s=\"http://www.w3.org/2003/05/soap-envelope\" xmlns:a=\"http://www.w3.org/2005/08/addressing\">";
    xml += "<s:Header>";
    xml += "<a:Action s:mustUnderstand=\"1\">http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue</a:Action>";
    xml += "<a:MessageID>urn:uuid:{" + str(uuid4()) + "}</a:MessageID>";
    xml += "<a:ReplyTo>";
    xml += "<a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>";
    xml += "</a:ReplyTo>";
    xml += "<Security s:mustUnderstand=\"1\" xmlns:u=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd\" xmlns=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd\">";
    xml += "<u:Timestamp  u:Id=\"{" + str(uuid4()) + "}\">";
    xml += "<u:Created>" + datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ') + "</u:Created>";
    xml += "<u:Expires>" + (datetime.datetime.now() + datetime.timedelta(hours=1)).strftime(
        '%Y-%m-%dT%H:%M:%S.%fZ') + "</u:Expires>";
    xml += "</u:Timestamp>";
    xml += "<UsernameToken u:Id=\"{" + str(uuid4()) + "}\">";
    xml += "<Username>" + username + "</Username>";
    xml += "<Password Type=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordText\">" + password + "</Password>";
    xml += "</UsernameToken>";
    xml += "</Security>";
    xml += "<a:To s:mustUnderstand=\"1\">" + usernamemixed + "</a:To>";
    xml += "</s:Header>";
    xml += "<s:Body>";
    xml += "<trust:RequestSecurityToken xmlns:trust=\"http://docs.oasis-open.org/ws-sx/ws-trust/200512\">";
    xml += "<wsp:AppliesTo xmlns:wsp=\"http://schemas.xmlsoap.org/ws/2004/09/policy\">";
    xml += "<a:EndpointReference>";
    xml += "<a:Address>" + urnaddress + "</a:Address>";
    xml += "</a:EndpointReference>";
    xml += "</wsp:AppliesTo>";
    xml += "<trust:RequestType>http://docs.oasis-open.org/ws-sx/ws-trust/200512/Issue</trust:RequestType>";
    xml += "</trust:RequestSecurityToken>";
    xml += "</s:Body>";
    xml += "</s:Envelope>";

    try:
        res = requests.post(usernamemixed, data=xml, headers={'Content-type': 'application/soap+xml; charset=UTF-8'})
    except requests.exceptions.RequestException as e:
        raise Exception(e)

    try:
        tokxml = etree.fromstring(res.text)
    except ImportError:
        raise Exception('Failed to import Token XML response')

    tok = tokxml.findall('.//xenc:EncryptedData', {'xenc': 'http://www.w3.org/2001/04/xmlenc#'})
    if len(tok) == 0:
        raise Exception('Failed to fetch token')
    token = etree.tostring(tok[0])

    return token.decode("utf-8")


import setuptools

with open('requirements.txt') as requirements:
    required = requirements.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pydynamics",
    version="0.0.3",
    author="Dan Goscomb",
    author_email="dan@flowplex.co.uk",
    description="Talk to On-Prem Dynamics CRM",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dangoscomb/pydynamics",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',
    install_requires=required
)